#include "Screen.H"
#include "Logger.H"
#include <curses.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>

#define FAKE_BLINK

extern Logger logger;

bool SCR_RESIZED=false;
void (*PreviousHandler)(int)=NULL;


Screen::Screen(CharColor fore, CharColor back):
	_grd(NULL),
	_width(0),
	_height(0) {
	resetTiming();
	pthread_mutex_init(&_scrlock,NULL); 
	pthread_mutex_init(&_inplock,NULL); 
#ifndef CURFLOO_NO_COLOR
	_usedColors=2;
	_color=true;
	for (int x=0; x < cc_last; x++)
		for (int y=0; y < cc_last; y++)
			_colorChart[x][y]=-1;
	_colorChart[cc_undef][cc_undef]=(_usedColors-1);
	_currentColor=_colorChart[fore][back]=(_usedColors-1);
	_foreDef=getNativeColor(fore);
	_backDef=getNativeColor(back);
#else
	_color=false;
#endif
}

Screen::~Screen() {
	cleanup();
	pthread_mutex_destroy(&_scrlock);
	pthread_mutex_destroy(&_inplock);
	puts("");
}

int Screen::init() {
	initscr();
	_width=COLS;
	_height=LINES;
	_grd=new Grid(_width,_height);
	noecho();
	//nodelay(stdscr,TRUE);
	//cbreak();
	raw();
	timeout(0);
	curs_set(0);
#ifndef CURFLOO_NO_COLOR
	start_color();
	init_pair(_currentColor,_foreDef,_backDef);
	color_set(_currentColor,NULL);
#endif
	clearScr();
	paint();
	PreviousHandler=signal(SIGWINCH,sigResizeHandler);
	pthread_create(&_inputReaderThr,NULL,(void *(*)(void *))inputReaderRunner,
		this);
	pthread_create(&_inputHandlerThr,NULL,(void *(*)(void *))inputHandlerRunner,
		this);
	pthread_create(&_updateThr,NULL,(void *(*)(void *))updateRunner, this);
	return 0;
}

int Screen::cleanup() {
	pthread_cancel(_inputReaderThr);
	pthread_cancel(_inputHandlerThr);
	pthread_cancel(_updateThr);
	if (_grd) {
		delete _grd;
		_grd=NULL;
	}
	bailout();
	return 0;
}

void Screen::bailout() {
	clear();
	refresh();
	endwin();
}

int Screen::colorOn() {
#ifndef CURFLOO_NO_COLOR
	_color=true;
	return 0;
#else
	return -1;
#endif
}

void Screen::colorOff() {
#ifndef CURFLOO_NO_COLOR
	_color=false;
	color_set(1,NULL);
#endif
}

int Screen::registerWindow(Window *win, const string name, unsigned int x,
		unsigned int y, Char activator) {
	return (_list.add(name,win,x,y,activator));
}

void Screen::clearScr() {
	Char
//		spc(0x20);
		spc('.');

	(*_grd).fill(spc);
}

int Screen::paint() {
	return (paint(0,0,(*_grd)));
}

int Screen::paint(unsigned int x, unsigned int y, const Grid &grd) {
	unsigned int
		xInc,
		yInc,
		idx;
	Char
		chr;
	char
		hack[]="%s", // the FSCKng proto for printw is wrong!
		buf[_width+1];

	if ((x>=_width) || (y>=_height)) {
		return -1;
	}
	pthread_mutex_lock(&_scrlock);
    for (yInc=0; ((y+yInc)<_height) && (yInc<grd.height()) && (!SCR_RESIZED);
			yInc++) {
		move((y+yInc),x);
        for (xInc=0; ((x+xInc)<_width) && (xInc<grd.width()) && (!SCR_RESIZED);
				xInc++) {
			chr=grd.get(xInc,yInc);
			setScreenByChar(chr);
			idx=0;
			while ((chr.isLike(grd.get(xInc,yInc))) && ((x+xInc) < _width)) {
				if (grd.get(xInc++,yInc).isPrintable())
					buf[idx++]=grd.get(xInc-1,yInc).getSymbol();
				else
					buf[idx++]=0x20;
			}
			xInc--;
			buf[idx]=0;
			if (SCR_RESIZED)
				break;
			printw(hack,buf);	
		}
	}
	pthread_mutex_unlock(&_scrlock);
	if (SCR_RESIZED) {
		SCR_RESIZED=false;
		return -2;
	}
	else {
		refresh();
	}
	return 0;
}

int Screen::snapshot(const string &fname, const Grid &grd) {
	unsigned int
		xInc,
		yInc;
	Char
		chr;
	char
		buf[grd.width()+1];
	FILE
		*fp;

	if ((fp=fopen(fname.c_str(),"w"))==NULL)
		return -1;
    for (yInc=0; yInc < grd.height(); yInc++) {
        for (xInc=0; xInc < grd.width(); xInc++) {
			chr=grd.get(xInc,yInc);
			if (!chr.isPrintable())
				chr=0x20;
			buf[xInc]=chr.getSymbol();
		}
		buf[xInc]=0;
		fprintf(fp,"%s\n",buf);
	}
	fclose(fp);
	return 0;
}

int Screen::adjustTiming(unsigned long rwait, unsigned long iwait,
		unsigned long ihwait) {
	_refresh_wait=rwait;
	_input_wait=iwait;
	_input_handler_wait=ihwait;
	_timer_hack=TIMER_HACK(_refresh_wait);
	_timer=0;
	return 0;
}

void Screen::getTiming(unsigned long &rwait, unsigned long &iwait,
		unsigned long &ihwait) {
	rwait=_refresh_wait;
	iwait=_input_wait;
	ihwait=_input_handler_wait;
}

void Screen::resetTiming() {
	_input_wait=D_INPUT_WAIT;
	_input_handler_wait=D_INPUT_HANDLER_WAIT;
	_refresh_wait=D_REFRESH_WAIT;
	_timer_hack=TIMER_HACK(_refresh_wait);
	_timer=0;
}

short Screen::getNativeColor(CharColor color) {
	switch(color) {
		case cc_black:
			return COLOR_BLACK;
		case cc_red:
			return COLOR_RED;
		case cc_green:
			return COLOR_GREEN;
		case cc_yellow:
			return COLOR_YELLOW;
		case cc_blue:
			return COLOR_BLUE;
		case cc_magenta:
			return COLOR_MAGENTA;
		case cc_cyan:
			return COLOR_CYAN;
		case cc_white:
			return COLOR_WHITE;
		default:
			return -1;
	}
}

int Screen::setScreenByChar(Char chr) {
	short
		colorp,
		fore,
		back;

	if (chr.isBold()) {
		attron(A_BOLD);
	}
	else {
		attroff(A_BOLD);
	}
	if (chr.isUnderline()) {
		attron(A_UNDERLINE);
	}
	else {
		attroff(A_UNDERLINE);
	}
	if (chr.isBlink()) {
#ifdef FAKE_BLINK
		// fuckin blink dont seem to be workin...
		// so now we do dees dee ol fashioned way
		if(_timer  >= _timer_hack) {
			chr.tReverse();
		}
#else													//FAKE_BLINK
		attron(A_BLINK);
#endif
	}
#ifndef FAKE_BLINK
	else {
		attroff(A_BLINK);
	}
#endif													// FAKE_BLINK
	if (chr.isReverse()) {
		attron(A_REVERSE);
	}
	else {
		attroff(A_REVERSE);
	}
	if (chr.isStandout()) {
		attron(A_STANDOUT);
	}
	else {
		attroff(A_STANDOUT);
	}
#ifndef CURFLOO_NO_COLOR
	if ((_color) &&
		((colorp=_colorChart[chr.getFore()][chr.getBack()])!= _currentColor)) {
		if (colorp==-1) {
			colorp=_usedColors++;
			_colorChart[chr.getFore()][chr.getBack()]=colorp;
			if ((fore=getNativeColor(chr.getFore())) == -1)
				fore=_foreDef;
			if ((back=getNativeColor(chr.getBack())) == -1)
				back=_backDef;
			init_pair(colorp,fore,back);
		}
		color_set(colorp,NULL);
		_currentColor=colorp;
	}
#endif
	return 0;
}

Char Screen::getInput() {
	int
		key;
	bool
		haderr=true;
	while (1) {
		key=getch();
		if (key==ERR) {
			if (!haderr) {
				haderr=true;
				pthread_mutex_lock(&_inplock);
				_inp.push(key);
				pthread_mutex_unlock(&_inplock);
			}
			usleep(_input_wait);
			continue;     
		}
		haderr=false;
		pthread_mutex_lock(&_inplock);
		_inp.push(key);
		pthread_mutex_unlock(&_inplock);
	}
}

void *Screen::inputReaderRunner(Screen *this_) {
	while(1) {
		this_->getInput();
	}
}

int Screen::handleInput(Char key) {
	Widget
		*wp;

	if (_list.size()==0)
		return -1;
	for (int i=0; i < _list.size(); i++) {
		wp=_list.get();	
		if ((*wp).input(key)==ws_accepted) {
			return 0;	
		}
		else {
		}
	}
	if (key.isInterrupt()) {
		raise(SIGINT);
	}
	else if (key.isF1()) {
		_list.toBack(0);
	}
	else if (key.isUpArrow()) {
		_list.moveBy(0,0,-1);
	}
	else if (key.isDownArrow()) {
		_list.moveBy(0,0,+1);
	} 
	else if (key.isRightArrow()) {
		_list.moveBy(0,+1,0);
	}
	else if (key.isLeftArrow()) {
		_list.moveBy(0,-1,0);
	}
	return 0;
}

void *Screen::inputHandlerRunner(Screen *this_) {
	Char
		c;
	int
		inp[KEY_SEQUENCE_LENGTH],
		key,
		ctrl;	

	while(1) {
		if (!(this_->_inp.size())) {
			usleep(this_->_input_handler_wait);
			continue;
		}
		ctrl=0;
		memset(&inp,0,KEY_SEQUENCE_LENGTH*(sizeof (int)));
		pthread_mutex_lock(&(this_->_inplock));
		key=this_->_inp.front();
		pthread_mutex_unlock(&(this_->_inplock));
		while ((key != ERR) && (ctrl < (KEY_SEQUENCE_LENGTH-1))) {
			inp[ctrl++]=key;
			if ((ctrl==1) && (key >= 0x20) && (key < 0x7F))
				break;
			pthread_mutex_lock(&(this_->_inplock));
			this_->_inp.pop();
			key=this_->_inp.front();
			pthread_mutex_unlock(&(this_->_inplock));
		}
		pthread_mutex_lock(&(this_->_inplock));
		this_->_inp.pop();
		pthread_mutex_unlock(&(this_->_inplock));
		inp[ctrl]=0;
		if  (ctrl)
			this_->handleInput(this_->keymap(inp));
	}
}

int Screen::update() {
	Widget
		*wp;
	Char
		activator;
	int
		i,
		x,
		y;
	string
		name;
	bool
		winmv=false,
		cng=false,
		mvd=false,
		vis=true;

	if (_list.size()) {
		for (i=_list.size(); i; i--) {
			wp=_list.get(i-1,name,x,y,mvd,vis,activator);	
			if (mvd) {
				clearScr();
				break;
			}	
		}
		for (i=_list.size(); i; i--) {
			wp=_list.get(i-1,name,x,y,winmv,vis,activator);	
			if (!vis)
				continue;
			if ((mvd) || (cng)) {
				(*_grd).set(x,y,(*wp).draw());
				if (winmv)
					_list.moved(i-1);
			}
			else if ((*wp).update()) {
				(*_grd).set(x,y,(*wp).draw());
				cng=true;
			}
		}
		if (mvd || cng)	{
			if (paint() == -2) {
				delete(_grd);
				//sigResizeHandler(SIGWINCH); //this is just to be sure
				_width=COLS;
				_height=LINES;
				_grd=new Grid(_width,_height);
				clearScr();
			}
		}
	}
	usleep(_refresh_wait);
	return (!(mvd || cng));
}

void *Screen::updateRunner(Screen *this_) {
	while(1) {
		this_->update();
		if (this_->_timer++ >= ((this_->_timer_hack)*2))
			this_->_timer=0;
	}
}

void Screen::sigResizeHandler(int sig) {
	if (sig == SIGWINCH) {
		if (PreviousHandler) 
			PreviousHandler(sig);
		SCR_RESIZED=true;
	}
	else
		logger.log("heh.  wasnt it: %d (%d)",sig,SIGWINCH);
	logger.log("COLS=%d LINES=%d",COLS,LINES);
	signal(SIGWINCH,sigResizeHandler);
}
